﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using System.Windows.Media;
using Buddy.Coroutines;
using ff14bot;
using ff14bot.AClasses;
using ff14bot.Behavior;
using ff14bot.BotBases;
using ff14bot.Enums;
using ff14bot.Helpers;
using ff14bot.Managers;
using ff14bot.Navigation;
using ff14bot.NeoProfiles;
using ff14bot.Objects;
using ff14bot.RemoteWindows;
using TreeSharp;

namespace DestinyPlugin
{
    public class Destiny : BotPlugin
    {
        #region Fields

        private Composite _changeJobCoroutine;
        private Composite _changeLocationCoroutine;
        private Gui _settingsForm;

        #endregion

        #region Properties

        private static bool NeedToChangeJobs { get; set; }

        private static bool NeedToChangeLocation
        {
            get
            {
                return !FateManager.WithinFate && WorldManager.ZoneId != WorldManager.GetZoneForAetheryteId((uint)ZoneLevels[Core.Me.ClassLevel]);
            }
        }

        #endregion

        #region Overrides

        #region Required

        public override string Author
        {
            get { return "Wheredidigo"; }
        }

        // ReSharper disable once InconsistentNaming
        private const string _name = "Destiny";

        public override string Name
        {
            get { return _name; }
        }

        internal static string InternalName
        {
            get { return _name + " V" + _version; }
        }

        // ReSharper disable once InconsistentNaming
        private static readonly Version _version = new Version(2, 0, 0, 1);

        public override Version Version
        {
            get { return _version; }
        }

        internal static Version InternalVersion
        {
            get { return _version; }
        }

        #endregion

        #region Optional

        public override void OnInitialize()
        {
            NeedToChangeJobs = true;
            _changeJobCoroutine = new Decorator(j => BotManager.Current.EnglishName == "Fate Bot" && NeedToChangeJobs, new ActionRunCoroutine(j => ChangeJob()));
            _changeLocationCoroutine = new Decorator(l => BotManager.Current.EnglishName == "Fate Bot" && NeedToChangeLocation, new ActionRunCoroutine(l => ChangeZone()));
        }

        public override void OnEnabled()
        {
            TreeRoot.OnStart += OnBotStart;
            TreeRoot.OnStop += OnBotStop;
            TreeHooks.Instance.OnHooksCleared += OnOnHooksCleared;
            GameEvents.OnLevelUp += OnOnLevelUp;

            if (TreeRoot.IsRunning)
            {
                AddHooks();
            }
            SetFateBotSettings();
        }

        public override void OnDisabled()
        {
            TreeRoot.OnStart -= OnBotStart;
            TreeRoot.OnStop -= OnBotStop;
            TreeHooks.Instance.OnHooksCleared -= OnOnHooksCleared;
            GameEvents.OnLevelUp -= OnOnLevelUp;
            RemoveHooks();
        }

        public override void OnShutdown()
        {
            TreeRoot.OnStart -= OnBotStart;
            TreeRoot.OnStop -= OnBotStop;
            TreeHooks.Instance.OnHooksCleared -= OnOnHooksCleared;
            GameEvents.OnLevelUp -= OnOnLevelUp;
            RemoveHooks();
        }

        public override bool WantButton
        {
            get { return true; }
        }

        public override void OnButtonPress()
        {
            if (_settingsForm == null || _settingsForm.IsDisposed || _settingsForm.Disposing)
                _settingsForm = new Gui();

            _settingsForm.ShowDialog();
        }
        #endregion

        #endregion

        #region Hook Logic

        private void AddHooks()
        {
            Logging.Write(Colors.Pink, string.Format("Adding Destiny V{0} Hooks", Version));
            NeedToChangeJobs = true;
            if (!string.IsNullOrWhiteSpace(Settings.Instance.PreferedRoutine))
            {
                RoutineManager.PreferedRoutine = Settings.Instance.PreferedRoutine;
            }
            TreeHooks.Instance.InsertHook("TreeStart", 0, _changeJobCoroutine);
            TreeHooks.Instance.InsertHook("TreeStart", 1, _changeLocationCoroutine);
        }

        private void RemoveHooks()
        {
            Logging.Write(Colors.Pink, string.Format("Removing Destiny V{0} Hooks", Version));
            TreeHooks.Instance.RemoveHook("TreeStart", _changeJobCoroutine);
            TreeHooks.Instance.RemoveHook("TreeStart", _changeLocationCoroutine);
        }

        #endregion

        #region Event Handlers

        private void OnBotStop(BotBase bot)
        {
            RemoveHooks();
        }

        private void OnBotStart(BotBase bot)
        {
            AddHooks();
        }

        private void OnOnHooksCleared(object sender, EventArgs e)
        {
            AddHooks();
        }

        public void OnOnLevelUp(object sender, EventArgs e)
        {
            if (Core.Me.Levels.Where(
                    x => (x.Key.IsDisciplesOfMagic() | x.Key.IsDisciplesOfWar()) && GearSets[x.Key] != 0)
                    .OrderBy(x => x.Value)
                    .FirstOrDefault()
                    .Value <= (Core.Me.ClassLevel - 5))
            {
                NeedToChangeJobs = true;
            }
            SetFateBotSettings();
        }

        #endregion

        #region Coroutine Logic

        private static async Task<bool> ChangeJob()
        {
            if (Core.Me.InCombat)
            {
                Logging.Write(string.Format("[{0}] We are currently in Combat, waiting until out of combat to change jobs.", InternalName));
                return false;
            }

            var job = Core.Me.Levels.Where(x => (x.Key.IsDisciplesOfMagic() | x.Key.IsDisciplesOfWar()) && GearSets[x.Key] != 0).OrderBy(x => x.Value).FirstOrDefault().Key;
            if (job.BaseClass() == Core.Me.CurrentJob.BaseClass())
            {
                NeedToChangeJobs = false;
                return true;
            }

            Logging.Write(string.Format("[{0}] Trying to change job to {1}", InternalName, job));
            ChatManager.SendChat(string.Format("/gs change {0}", GearSets[job]));
            await Coroutine.Wait(10000, () => Core.Me.CurrentJob == job);
            if (Core.Me.CurrentJob != job)
            {
                Logging.Write(string.Format("[{0}] Unsuccessful in trying to change job to {1}. Will try again.", InternalName, job));
                return false;
            }
            Logging.Write(string.Format("[{0}] Successfully changed job to {1}", InternalName, job));
            SetFateBotSettings();
            NeedToChangeJobs = false;
            return true;
        }

        private static async Task<bool> ChangeZone()
        {
            if (Core.Me.InCombat)
            {
                Logging.Write(string.Format("[{0}] We are currently in Combat, waiting until out of combat to change jobs.", InternalName));
                return false;
            }

            Logging.Write(string.Format("[{0}] We're teleporting to {1}.", InternalName, WorldManager.LocationKey[(uint)ZoneLevels[Core.Me.ClassLevel]]));
            Poi.Clear(string.Format("[{0}] We're teleporting to another zone.", InternalName));
            Navigator.Stop();
            await CarefulSleep(200);
            if (Core.Me.IsMounted)
            {
                Logging.Write(string.Format("[{0}] Dismounting so we can teleport to a new zone.", InternalName));
                await CommonTasks.StopAndDismount();
            }
            WorldManager.TeleportById((uint)ZoneLevels[Core.Me.ClassLevel]);
            await CarefulSleep(1000);

            while (CanTeleport && !Core.Me.IsCasting)
            {
                if (Core.Me.IsMounted) { await CommonTasks.StopAndDismount(); }
                WorldManager.TeleportById((uint)ZoneLevels[Core.Me.ClassLevel]);
                await CarefulSleep(1000);
            }

            await CarefulSleep(15000);

            if (Settings.Instance.SetAetheryteAsHome)
            {
                await SetAetheryteAsHome();
            }

            return true;
        }

        private static async Task<bool> SetAetheryteAsHome()
        {
            var aetheryte = GameObjectManager.GetObjectByNPCId((uint)ZoneLevels[Core.Me.ClassLevel]);
            if (aetheryte != null)
            {
                while (!SelectString.IsOpen)
                {
                    aetheryte.Interact();
                    await CarefulSleep(1000);
                }
                while (!SelectYesno.IsOpen)
                {
                    SelectString.ClickSlot(0);
                    await CarefulSleep(1000);
                }
                
                SelectYesno.ClickYes();
            }

            return true;
        }

        private static void SetFateBotSettings()
        {
            FatebotSettings.Instance.BossEnabled = Settings.Instance.BossEnabled;
            FatebotSettings.Instance.EscortEnabled = Settings.Instance.EscortEnabled;
            FatebotSettings.Instance.MonsterSlayingEnabled = Settings.Instance.MonsterSlayingEnabled;
            FatebotSettings.Instance.LevelCheck = Settings.Instance.LevelCheck;
            FatebotSettings.Instance.MaxLevel = Settings.Instance.MaxLevel;
            FatebotSettings.Instance.MinLevel = Core.Me.ClassLevel - Settings.Instance.MinLevel;
            FatebotSettings.Instance.BossPercentRequired = Settings.Instance.BossPercentRequired;
        }
        #endregion

        #region Helpers

        private static Dictionary<int, int> _zoneLevels;

        internal static Dictionary<int, int> ZoneLevels
        {
            get { return _zoneLevels ?? (_zoneLevels = BuildZoneLevels()); }
            set { _zoneLevels = value; }
        }

        private static Dictionary<int, int> BuildZoneLevels()
        {
            return new Dictionary<int, int>
            {
                {1, Settings.Instance.Levels01_10Location},
                {2, Settings.Instance.Levels01_10Location},
                {3, Settings.Instance.Levels01_10Location},
                {4, Settings.Instance.Levels01_10Location},
                {5, Settings.Instance.Levels01_10Location},
                {6, Settings.Instance.Levels01_10Location},
                {7, Settings.Instance.Levels01_10Location},
                {8, Settings.Instance.Levels01_10Location},
                {9, Settings.Instance.Levels01_10Location},
                {10, Settings.Instance.Levels10_15Location},
                {11, Settings.Instance.Levels10_15Location},
                {12, Settings.Instance.Levels10_15Location},
                {13, Settings.Instance.Levels10_15Location},
                {14, Settings.Instance.Levels10_15Location},
                {15, Settings.Instance.Levels15_20Location},
                {16, Settings.Instance.Levels15_20Location},
                {17, Settings.Instance.Levels15_20Location},
                {18, Settings.Instance.Levels15_20Location},
                {19, Settings.Instance.Levels15_20Location},
                {20, Settings.Instance.Levels20_25Location},
                {21, Settings.Instance.Levels20_25Location},
                {22, Settings.Instance.Levels20_25Location},
                {23, Settings.Instance.Levels20_25Location},
                {24, Settings.Instance.Levels20_25Location},
                {25, Settings.Instance.Levels25_30Location},
                {26, Settings.Instance.Levels25_30Location},
                {27, Settings.Instance.Levels25_30Location},
                {28, Settings.Instance.Levels25_30Location},
                {29, Settings.Instance.Levels25_30Location},
                {30, Settings.Instance.Levels30_35Location},
                {31, Settings.Instance.Levels30_35Location},
                {32, Settings.Instance.Levels30_35Location},
                {33, Settings.Instance.Levels30_35Location},
                {34, Settings.Instance.Levels30_35Location},
                {35, Settings.Instance.Levels35_40Location},
                {36, Settings.Instance.Levels35_40Location},
                {37, Settings.Instance.Levels35_40Location},
                {38, Settings.Instance.Levels35_40Location},
                {39, Settings.Instance.Levels35_40Location},
                {40, Settings.Instance.Levels40_45Location},
                {41, Settings.Instance.Levels40_45Location},
                {42, Settings.Instance.Levels40_45Location},
                {43, Settings.Instance.Levels40_45Location},
                {44, Settings.Instance.Levels40_45Location},
                {45, Settings.Instance.Levels45_50Location},
                {46, Settings.Instance.Levels45_50Location},
                {47, Settings.Instance.Levels45_50Location},
                {48, Settings.Instance.Levels45_50Location},
                {49, Settings.Instance.Levels45_50Location},
                {50, Settings.Instance.Levels45_50Location}
            };
        }

        private static Dictionary<ClassJobType, int> _gearSets;

        internal static Dictionary<ClassJobType, int> GearSets
        {
            get { return _gearSets ?? (_gearSets = BuildGearSets()); }
            set { _gearSets = value; }
        }

        private static Dictionary<ClassJobType, int> BuildGearSets()
        {
            return new Dictionary<ClassJobType, int>
            {
                {ClassJobType.Archer, Settings.Instance.Archer_BardGearSet},
                {ClassJobType.Bard, Settings.Instance.Archer_BardGearSet},
                {ClassJobType.Gladiator, Settings.Instance.Gladiator_PaladinGearSet},
                {ClassJobType.Paladin, Settings.Instance.Gladiator_PaladinGearSet},
                {ClassJobType.Lancer, Settings.Instance.Lancer_DragoonGearSet},
                {ClassJobType.Dragoon, Settings.Instance.Lancer_DragoonGearSet},
                {ClassJobType.Marauder, Settings.Instance.Marauder_WarriorGearSet},
                {ClassJobType.Warrior, Settings.Instance.Marauder_WarriorGearSet},
                {ClassJobType.Pugilist, Settings.Instance.Pugilist_MonkGearSet},
                {ClassJobType.Monk, Settings.Instance.Pugilist_MonkGearSet},
                {ClassJobType.Rogue, Settings.Instance.Rogue_NinjaGearSet},
                {ClassJobType.Ninja, Settings.Instance.Rogue_NinjaGearSet},
                {ClassJobType.Conjurer, Settings.Instance.Conjurer_WhiteMageGearSet},
                {ClassJobType.WhiteMage, Settings.Instance.Conjurer_WhiteMageGearSet},
                {ClassJobType.Thaumaturge, Settings.Instance.Thaumaturge_BlackMageGearSet},
                {ClassJobType.BlackMage, Settings.Instance.Thaumaturge_BlackMageGearSet},
                {ClassJobType.Arcanist, Settings.Instance.ArcanistGearSet},
                {ClassJobType.Summoner, Settings.Instance.SummonerGearSet},
                {ClassJobType.Scholar, Settings.Instance.ScholarGearSet}
            };
            /*      Ready for Heavensward
            {ClassJobType.Astrologian, Settings.Instance.AstrologianGearSet},
            {ClassJobType.DarkKnight, Settings.Instance.DarkKnightGearSet},
            {ClassJobType.Machinist, Settings.Instance.MachinistGearSet}
            */
        }

        private static async Task<bool> CarefulSleep(uint ms)
        {
            var stopwatch = new Stopwatch();
            stopwatch.Start();

            while (stopwatch.ElapsedMilliseconds < ms && CanTeleport)
            {
                await Coroutine.Yield();
            }
            stopwatch.Stop();

            return CanTeleport;
        }

        private static bool CanTeleport
        {
            get
            {
                return !Core.Me.IsDead
                       && !Core.Me.InCombat
                       && !FateManager.WithinFate;
            }
        }
        #endregion
    }
}